/*
** htree.c
** CS216 Spring 2006
** Problem Set 4
*/

# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include <assert.h>
# include "htree.h"

# define MAX_CHAR 128

/*
** Some useful utility routines
*/

static int power (int base, int exp) {
	int res = 1;
	while (exp > 0) {
		res *= base;
		exp--;
	}

	return res;
}

static char *showBits (unsigned char c) {
	char *res = (char *) malloc (sizeof (*res) * 9);
	int i;

	if (res == NULL) {
	  fprintf (stderr, "Error: cannot allocate memory in showBits");
	  exit (EXIT_FAILURE);
	}

	for (i = 0; i < 8; i++) {
		unsigned char mask = (unsigned char) power (2, (7 - i));
		res[i] = (c & mask) ? '1' : '0';
	}
	res[8] = '\0';
	return res;
}

/*
** htree implementation
*/

int htree_isLeaf (htree t) {
	return (t->left == NULL && t->right == NULL);
}

static void htree_print_indent (htree t, int indent) {
	if (t == NULL) {
		return;
	} else {
		int i;
		for (i = 0; i < indent; i++) {
			printf (" ");
		}
		printf ("%d (%d)\n", (int) t->letter, t->count);
		htree_print_indent (t->left, indent + 2);
		htree_print_indent (t->right, indent + 2);
	}
}

void htree_print (htree t) {
	htree_print_indent (t, 0);
}

static char *findClosingBrace (char *s) {
	int depth = 1; /* already had one open */
	char *ts = s;

	while (*ts != '\0') {
		if (*ts == '[') {
			depth++;
		} else if (*ts == ']') {
			depth--;
			if (depth == 0) {
				break; /* found the closing ] */
			}
		}
		ts++;
	}

	if (*ts != ']') {
		fprintf (stderr, "Error: no closing brace: %s\n", s);
		exit (EXIT_FAILURE);
	}

	return ts;
}

htree htree_parse (char *s) {
	assert (s != NULL);

	if (*s == '.') {
		return NULL;
	} else if (*s == '[') {
		/* find matching ']' */
		char *ts = s + 1;
		char *left = ts;
		char *right;
		htree tleft;
		htree tright;
		htree res;

		ts = findClosingBrace (ts);
		assert (*ts == ']');
		*ts = '\0'; /* temporarily for parsing */
		tleft = htree_parse (left);
		*ts = ']';
		ts++;
		assert (*ts == '[');
		ts++;
		right = ts;
		ts = findClosingBrace (right);
		assert (*ts == ']');
		*ts = '\0';
		tright = htree_parse (right);
		*ts = ']';

		res = (htree) (malloc (sizeof (*res)));
		if (res == NULL) {
		  fprintf (stderr, "Error: cannot allocate memory in htree_parse");
		  exit (EXIT_FAILURE);
		}

		res->left = tleft;
		res->right = tright;
		if (tleft != NULL) {
			tleft->parent = res;
		}
		if (tright != NULL) {
			tright->parent = res;
		}
		res->parent = NULL;
		res->letter = '\0';
		res->count = 0;
		return res;
	} else {
		/* should be a decimal number */
		htree res;
		int val;
		char *ts = s;
		while (*ts != '\0' && *ts != '\n') {
			if (*ts < '0' || *ts > '9') {
				fprintf (stderr, "Error: non-numeric in node: [%d] %s", *ts, s);
				exit (EXIT_FAILURE);
			}
			ts++;
		}

		val = atoi (s);
		assert (val >= 0 && val < MAX_CHAR);
		res = (htree) (malloc (sizeof (*res)));
		if (res == NULL) {
		  fprintf (stderr, "Error: cannot allocate memory in htree_parse");
		  exit (EXIT_FAILURE);
		}
		res->left = NULL;
		res->right = NULL;
		res->parent = NULL;
		res->letter = val;
		res->count = 0;
		return res;
	}
}

/* post: returns a machine-readable string representation of t
** note: htree_parse (htree_unparse (t)) should equal t for all possible htree objects t.
**
** This procedure leaks memory.  (See question 4.)
*/

char *htree_unparse (htree t) {
	char *res;

	if (t == NULL) {
		return ".";
	}
	else if (htree_isLeaf (t)) { /* reached a leaf */
		res = (char *) malloc (4 * sizeof (*res));
		/* This is a Windows non-standard function.  On standard C implementations
		** (such as most Unix systems), use snprintf instead of _snprintf.
		*/
		_snprintf (res, 4, "%d", t->letter); /* print as a number (ASCII code) */
		return res;
	} else {
		char *tl = htree_unparse (t->left);
		char *tr = htree_unparse (t->right);
		/* Allocate space for concatenations, [] characters, and null */
		res = (char *) malloc (sizeof (*res) * (strlen(tl) + strlen (tr) + 5));
		if (res == NULL) {
		  fprintf (stderr, "Error: cannot allocate memory in htree_unparse");
		  exit (EXIT_FAILURE);
		}
		strcpy (res, "[");
		strcat (res, tl);
		strcat (res, "][");
		strcat (res, tr);
		strcat (res, "]");
		return res;
	}
}

/*
** pre: s is a null-terminated string
** post: returns the Huffman tree for optimally encoding s 
*/

htree htree_buildTree (char *s) {
	/*
	** Implement this for Question 5
	*/

	/* This is a stub implementation that is very non-optimal 
	** (it constructs a valid encoding tree, but does not consider
	** the letter frequency at all)
	*/

	int i;
	htree tnodes[MAX_CHAR];
	int numnodes = 0;

	for (i = 0; i < MAX_CHAR; i++) {
		htree newnode = (htree) malloc (sizeof (*newnode));
		if (newnode == NULL) {
			fprintf (stderr, "Unable to allocate: %u", sizeof (*newnode));
			exit (EXIT_FAILURE);
		}
		newnode->left = NULL;
		newnode->right = NULL;
		newnode->letter = (char) i;
		newnode->count = 1;
		newnode->parent = NULL;
		tnodes[numnodes++] = newnode;
	}

	/*
	** Now, we have all the leaves, we need to form the tree by
	** merging nodes.
	*/

	while (numnodes > 1) {
		i = 0;
		while (i < numnodes) {
			htree newnode = (htree) malloc (sizeof (*newnode));
			if (newnode == NULL) {
				fprintf (stderr, "Unable to allocate: %u", sizeof (*newnode));
				exit (EXIT_FAILURE);
			}
			newnode->left = tnodes[i];
			newnode->right = tnodes[i + 1];
			tnodes[i]->parent = newnode;
			tnodes[i+1]->parent = newnode;
			newnode->letter = '\0';
			newnode->count = 1;
			newnode->parent = NULL;
			tnodes[i] = newnode;
			tnodes[i + 1] = tnodes[numnodes - 1];
			numnodes--;
			i++;
		}
	}

	/* useful for debugging: htree_print (tnodes[0]); */
	return tnodes[0];
}

char *htree_encodeChar (htree h, char c)
{
	if (h == NULL) {
		return NULL;
	} else if (htree_isLeaf (h)) {
		if (h->letter == c) {
			int depth = 0;
			char *res;
			htree ht = h;
			while (ht->parent != NULL) {
				depth++;
				ht = ht->parent;
			}

			res = (char *) malloc (sizeof (*res) * (depth + 1));
			if (res == NULL) {
			  fprintf (stderr, "Error: cannot allocate memory in htree_encodeChar");
			  exit (EXIT_FAILURE);
			}

			res[depth] = '\0';
			ht = h->parent;
			while (depth > 0) {
				if (ht->left == h) {
					res[depth - 1] = '0';
				} else if (ht->right == h) {
					res[depth - 1] = '1';
				} else {
					fprintf (stderr, "Error! Bad tree!");
					exit (EXIT_FAILURE);
				}

				h = ht;
				ht = ht->parent;
				depth--;
			}
			return res;
		} else {
			return NULL;
		}
	} else {
		char *res = htree_encodeChar (h->left, c);

		if (res == NULL) {
			res = htree_encodeChar (h->right, c);
		}

		return res;
	}
}

void htree_encodeChars (char *s, FILE *outfile) {
	/* first, output the htree encoding */
	htree h;	    
	
	h = htree_buildTree (s);
 	fprintf (outfile, "%s\n", htree_unparse (h));

	while (*s != '\0') {
		char c = *s++;
		char *bits = htree_encodeChar (h, c);
		fprintf (outfile, "%s", bits);
	}
}

/* This is just a guess which is an overestimate...could figure it out, but space is cheap */
# define MAX_TREE_LINE (MAX_CHAR * 12)
# define BLOCK_SIZE 1024

char *htree_decodeChars (FILE *infile) {
	/* The first line of the file is the tree */
	char *s = (char *) malloc (sizeof (*s) * MAX_TREE_LINE);
	htree h;
	htree curnode;
	int inchar;
	int allocsize = BLOCK_SIZE;
	char *res = (char *) malloc (sizeof (*res) * allocsize);
	char *tres = res;

	if (s == NULL || res == NULL) {
	  fprintf (stderr, "Error: cannot allocate memory in htree_decodeChars");
	  exit (EXIT_FAILURE);
	}

	*tres = '\0';

	s = fgets (s, MAX_TREE_LINE, infile);
	if (s == NULL) {
		fprintf (stderr, "Error: cannot read first line of file.\n");
		exit (EXIT_FAILURE);
	} 
	h = htree_parse (s);
	free (s);
	curnode = h;

	while ((inchar = fgetc (infile)) != EOF) {
		char c = (char) inchar;
		if (c == '\0') {
			fprintf (stderr, "Error: read file has null character!\n");
			exit (EXIT_FAILURE);
		}

		if (c == '0') { /* go left */
			/* Complete this code */
		} else if (c == '1') { /* go right */
			/* Complete this code */
		} else {
			fprintf (stderr, "Error: bad character: %c [%d]\n%s\n", c, (int) c, res);
			exit (EXIT_FAILURE);
		}

		if (htree_isLeaf (curnode)) {
			/*
			** Complete this code...think about what to do when you reach a leaf.
			** Remember to be careful you have enough space before adding characters to res.
			*/			

			curnode = h;
		}
	}

	*tres = '\0';
	assert (curnode == h); /* Check we are not in the middle of a traversal */
	return res;
}

void htree_encodeBits (char *s, FILE *outfile) {
	/* first, output the htree encoding */
	htree h;	    
	
	h = htree_buildTree (s);
 	fprintf (outfile, "%s\n", htree_unparse (h));

	/* Question 10: Complete this implementation */

	fclose (outfile);
}

char *htree_decodeBits (FILE *infile) {
	/* The first line of the file is the tree */
	char *s = (char *) malloc (sizeof (*s) * MAX_TREE_LINE);
	htree h;
	htree curnode;
	int allocsize = BLOCK_SIZE;
	char *res = (char *) malloc (sizeof (*res) * allocsize);
	char *tres = res;
	int used = 0;
	int nextchar;
	int curchar;
	int nextnextchar;

	if (s == NULL || res == NULL) {
	  fprintf (stderr, "Error: cannot allocate memory in htree_decodeBits");
	  exit (EXIT_FAILURE);
	}

	*tres = '\0';

	s = fgets (s, MAX_TREE_LINE, infile);

	if (s == NULL) {
		fprintf (stderr, "Error: cannot read first line of file.\n");
		exit (EXIT_FAILURE);
	} 

	h = htree_parse (s);
	free (s);
	curnode = h;

	curchar = fgetc (infile);
	assert (curchar != EOF);
	nextchar = fgetc (infile);
	assert (curchar != EOF);
	nextnextchar = fgetc (infile);

	while (1) {
		unsigned char c = (unsigned char) curchar;
		int i = 0;
		int numbits = 8;

		if (nextnextchar == EOF) {
			/*
			** now, nextchar encodes the number of valid bits in curchar
			*/
			short nvalid = (short) nextchar;
			assert (nvalid > 0 && nvalid <= 8);
			numbits = nvalid;
		}

		for (i = 0; i < numbits; i++) { /* read each bit */
			unsigned char bit;
			bit = c /* Question 10: Complete this expression to select the ith bit of c */ ;

			if (!bit) { 
				curnode = curnode->left;
				assert (curnode != NULL);
			} else { 
				curnode = curnode->right;
				assert (curnode != NULL);
			}

			if (htree_isLeaf (curnode)) {
				if (used + 1 > allocsize) {
					allocsize += BLOCK_SIZE;
					res = realloc (res, sizeof(*res) * allocsize);
					if (res == NULL) {
						fprintf (stderr, "Error: cannot allocate enough memory (%d)\n", 
							allocsize);
						exit (EXIT_FAILURE);
					}
					tres = res + used;
				}

				*tres++ = curnode->letter;
				used++;
				curnode = h;
			}
		}

		if (nextnextchar == EOF) {
			break;
		}

		curchar = nextchar;
		nextchar = nextnextchar;
		nextnextchar = fgetc (infile);
	}

	*tres = '\0';
	assert (curnode == h); /* Check we are not in the middle of a traversal */
	return res;
}
