/*
** huffman.c
** CS216 Spring 2006
** Problem Set 4
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "htree.h"

/*
** pre: f refers to a file that is open for reading
** post: returns a new string that represents all the characters in f
*/

# define BLOCK_SIZE 1024

char *readInputFile (FILE *f)
{
  int allocsize = BLOCK_SIZE;
  char *s = (char *) malloc (sizeof (*s) * allocsize);
  char *ts;
  int used = 0;
  int inchar;

  if (s == NULL) {
    fprintf (stderr, "Error: not enough memory to allocate s.");
    exit (EXIT_FAILURE);
  }

  ts = s;

  /*
  ** Read input file into string s.
  ** Be careful to always have enough space...
  */

  while ((inchar = fgetc (f)) != EOF) {
    /* We do not support null characters in read file */
    if (inchar == (int) '\0') {
      fprintf (stderr, "Error: read file has null character!\n");
      exit (EXIT_FAILURE);
    }

    if (used + 1 > allocsize) {
      allocsize += BLOCK_SIZE;
      s = realloc (s, sizeof(*s) * allocsize);
      ts = s + used;
    }

    *ts++ = (char) inchar;
    used++;
  }

  /* Add the null terminator */
  if (used + 1 > allocsize) {
    allocsize += BLOCK_SIZE;
    s = realloc (s, sizeof(*s) * allocsize);
    if (s == NULL) {
      fprintf (stderr, "Error: not enough memory to allocate: %d", allocsize);
      exit (EXIT_FAILURE);
    }
    ts = s + used;
  }

  *ts = '\0';
  return s;
}


/*
**
** Huffman takes two filenames as input, and stores the huffman encoded
** first file in the second file.
**
** e.g., huffman input.txt input.hfc
**
** will store the huffman-encoded version of input.txt in input.hfc.
**
** pre: 
**   The arguments are of the form 
**      [-d] [-b] <input> [<output>]
**   where input is a readable file and output (if provided) is a writable file.
*/

static void showUsage ()
{
  fprintf (stderr, "Usage: huffman [-d] [-b] <input> [<output>]\n");
  exit (EXIT_FAILURE);
}

int main(int argc, char **argv)
{
  char *inname;
  FILE *infile;
  char *outname;
  FILE *outfile;
  int argno = 1;
  int binarymode = 0;
  int decompress = 0;

  while (argno < argc && *argv[argno] == '-') {
    char *opt = argv[argno] + 1;
    while (*opt != '\0') {
      switch (*opt) {
      case 'b':
	binarymode = 1;
	break;
      case 'd':
	decompress = 1;
	break;
      default:
	showUsage ();
      }
      opt++;
    }
    argno++;
  }
  
  if (argno == argc) {
    showUsage ();
  }

  inname = argv[argno++];

  if (argno < argc) {
    outname = argv[argno++];
  } else {
    outname = NULL;
  }

  if (argno != argc) {
    showUsage ();
  }

  if (binarymode) {
    infile = fopen (inname, "rb"); 
    /* need to open in binary mode to handle Ctrl-Z correctly */ 
  } else { 
    infile = fopen (inname, "r"); 
  }  

  if (infile == NULL) {
    fprintf (stderr, "Cannot open input file for reading: %s\n", inname);
    exit (EXIT_FAILURE);
  }

  if (outname != NULL) {
    outfile = fopen (outname, "r"); 
    if (outfile != NULL) {
      fprintf (stderr, "Output file already exists.\n");
      exit (EXIT_FAILURE);
    }
    if (binarymode) {
      outfile = fopen (outname, "wb");
    } else {
      outfile = fopen (outname, "w");
    }
    if (outfile == NULL) {
      fprintf (stderr, "Cannot open output file: %s\n", outname);
      exit (EXIT_FAILURE);
    }
  } else {
    outfile = stdout;
  }

  if (decompress) {
    char *s;

    if (binarymode) {
      s = htree_decodeBits (infile);
    } else {
      s = htree_decodeChars (infile); 
    }

    fprintf (outfile, "%s", s);
    free (s);
  } else { /* we are encoding */
    char *s; 
    s = readInputFile (infile);

    if (binarymode) {
      htree_encodeBits (s, outfile);
    } else {
      htree_encodeChars (s, outfile); 
    }
    free (s);
  }

  return EXIT_SUCCESS;
}







